"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const nomatic_events_1 = require("nomatic-events");
const format = require("string-format");
class Logger extends nomatic_events_1.EventEmitter {
    constructor(name = 'root', options = {}) {
        super();
        this._levels = {
            trace: 50,
            debug: 40,
            info: 30,
            warn: 20,
            error: 10
        };
        this.transports = [];
        this.children = [];
        this.template = null;
        this.name = name;
        if (!options.levels) {
            options.levels = this.levels;
        }
        this.configure(options);
    }
    get levels() {
        return this._levels;
    }
    set levels(levels) {
        for (const level in this._levels) {
            if (!levels.hasOwnProperty(level) && this[level]) {
                delete this[level];
            }
        }
        for (const level in levels) {
            if (!this[level]) {
                this[level] = (messageOrData, data) => {
                    return this.log(level, messageOrData, data);
                };
            }
        }
        this._levels = levels;
        for (const child of this.children) {
            child.levels = levels;
        }
    }
    serialize(level, message, data = null) {
        const entry = {
            name: this.name,
            message: message,
            level: level,
            date: new Date()
        };
        if (data) {
            entry.data = data;
        }
        return entry;
    }
    configure(options) {
        if (options.transports) {
            this.transports = [];
            for (const i in options.transports) {
                this.use(options.transports[i]);
            }
        }
        if (options.template) {
            this.template = options.template;
        }
        if (options.levels) {
            this.levels = options.levels;
        }
        for (const child of this.children) {
            child.configure(options);
        }
    }
    create(name, options = {}) {
        if (!options.transports) {
            options.transports = this.transports;
        }
        if (!options.levels) {
            options.levels = this.levels;
        }
        const logger = new Logger(name, options);
        this.register(logger);
        return logger;
    }
    find(name) {
        for (const i in this.children) {
            if (this.children[i].name === name) {
                return this.children[i];
            }
        }
        return null;
    }
    log(level, messageOrData, data) {
        let message;
        if (!this.levels.hasOwnProperty(level)) {
            throw new Error('Invalid level: ' + level);
        }
        if (typeof messageOrData !== 'string') {
            if (!this.template) {
                throw new Error('`message` is not specified and `template` is not defined');
            }
            else {
                message = format(this.template, messageOrData);
                data = messageOrData;
            }
        }
        else {
            message = messageOrData;
        }
        const entry = this.serialize(level, message, data);
        this.emit(level, entry);
        this.emit('entry', entry);
        for (const transport of this.transports) {
            if (this.levels[entry.level] <= this.levels[transport.level]) {
                transport.execute(entry);
            }
        }
    }
    get(name) {
        let result = this.find(name);
        if (!result) {
            return this.create(name);
        }
        return result;
    }
    register(logger) {
        if (this.children.indexOf(logger) === -1) {
            for (const child of this.children) {
                if (child.name === logger.name) {
                    throw new Error('Child already exists with name "' + logger.name + '"');
                }
            }
            this.children.push(logger);
            return true;
        }
        return false;
    }
    use(transport) {
        if (this.transports.indexOf(transport) !== -1) {
            return;
        }
        if (transport.level && !this.levels[transport.level]) {
            throw new Error('Transport specifies invalid `level`: ' + transport.level);
        }
        this.transports.push(transport);
        for (const child of this.children) {
            child.use(transport);
        }
    }
}
exports.Logger = Logger;
exports.default = Logger;
//# sourceMappingURL=Logger.js.map